"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const deps_1 = require("../deps");
const fs = require("fs-extra");
const path = require("path");
const debug = require('debug')('cli:cache');
class PluginCache {
    constructor(config) {
        this.config = config;
        this.needsSave = false;
        this.fetchPromises = {};
    }
    async save() {
        await this.init();
        if (!this.needsSave)
            return;
        this.needsSave = false;
        this.saving = (async () => {
            debug('saving');
            if (!await this.canWrite()) {
                throw new Error('cache file modified, cannot save');
            }
            await deps_1.default.file.outputJSON(this.file, this.cache, { spaces: 0 });
        })();
    }
    async fetch(key, prop, fn) {
        await this.init();
        if (this.fetchPromises[key + prop])
            return this.fetchPromises[key + prop];
        return (this.fetchPromises[key + prop] = (async () => {
            let v = await this.get(key, prop);
            if (!v) {
                debug('fetching', key, prop);
                await this.set(key, prop, await fn());
            }
            return await this.get(key, prop);
        })());
    }
    async get(key, prop) {
        await this.init();
        if (!this.cache.plugins[key])
            return;
        return this.cache.plugins[key][prop];
    }
    async set(key, prop, v) {
        await this.init();
        if (!this.cache.plugins[key]) {
            this.cache.plugins[key] = {};
        }
        this.cache.plugins[key][prop] = v;
        this.needsSave = true;
        return this.cache.plugins[key][prop];
    }
    async reset(key) {
        await this.init();
        delete this.cache.plugins[key];
        this.fetchPromises = {};
        this.needsSave = true;
    }
    async init() {
        await this.saving;
        if (this._init)
            return this._init;
        return (this._init = (async () => {
            debug('init');
            this.cache = (await this.read()) || {
                version: this.config.version,
                plugins: {},
            };
        })());
    }
    get file() {
        return path.join(this.config.cacheDir, 'plugins.json');
    }
    async read() {
        try {
            this.mtime = await this.getLastUpdated();
            let cache = await fs.readJSON(this.file);
            if (cache.version !== this.config.version) {
                debug('cache version mismatch');
                return;
            }
            if (!cache.plugins)
                this.cache.plugins = {};
            return cache;
        }
        catch (err) {
            if (err.code === 'ENOENT') {
                debug(err);
            }
            else
                throw err;
        }
    }
    async canWrite() {
        if (!this.mtime)
            return true;
        return (await this.getLastUpdated()) === this.mtime;
    }
    async getLastUpdated() {
        try {
            const stat = await deps_1.default.file.stat(this.file);
            return stat.mtime.getTime();
        }
        catch (err) {
            if (err.code !== 'ENOENT')
                throw err;
        }
    }
}
exports.PluginCache = PluginCache;
