"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const cli_ux_1 = require("cli-ux");
const deps_1 = require("../deps");
const manager_1 = require("./manager");
const ts_lodash_1 = require("ts-lodash");
const path = require("path");
class Plugins extends manager_1.PluginManager {
    constructor({ config }) {
        super({ config });
        this.debug = require('debug')('cli:plugins');
        this.lock = new deps_1.default.Lock(config, path.join(config.cacheDir, 'plugins.lock'));
    }
    async install(options) {
        let downgrade = await this.lock.upgrade();
        let name = options.type === 'user' ? options.name : await this.getLinkedPackageName(options.root);
        if (!options.force && (await this.pluginType(name))) {
            throw new Error(`${name} is already installed. Run with --force to install anyway`);
        }
        if (options.type === 'link') {
            await this.link.install(options.root);
            await this.manifest.add({ type: options.type, name, root: options.root });
        }
        else {
            await this.user.install(name, options.tag);
            await this.manifest.add({ type: options.type, name, tag: options.tag });
        }
        await this.save();
        await downgrade();
    }
    async update() {
        let downgrade = await this.lock.upgrade();
        await this.migrate();
        await this.user.update();
        await this.save();
        await downgrade();
    }
    async uninstall(name) {
        let downgrade = await this.lock.upgrade();
        const type = await this.pluginType(name);
        if (!type) {
            const linked = await this.link.findByRoot(name);
            if (linked) {
                name = linked.name;
            }
            else
                throw new Error(`${name} is not installed`);
        }
        await this.manifest.remove(name);
        await this.save();
        if (type === 'user')
            await this.user.uninstall(name);
        await downgrade();
    }
    async init() {
        await super.init();
        if (await this.needsRefresh) {
            let downgrade = await this.lock.upgrade();
            cli_ux_1.default.action.start('Refreshing plugins');
            await this.refresh();
            await this.load;
            await this.save();
            cli_ux_1.default.action.stop();
            await downgrade();
        }
        await this.load;
    }
    async _init() {
        const submanagerOpts = { config: this.config, manifest: this.manifest, cache: this.cache, lock: this.lock };
        this.builtin = new deps_1.default.Builtin(submanagerOpts);
        this.submanagers.push(this.builtin);
        try {
            this.core = new deps_1.default.CorePlugins(submanagerOpts);
        }
        catch (err) {
            cli_ux_1.default.warn(err, { context: 'core plugins' });
        }
        if (this.config.userPlugins) {
            try {
                this.user = new deps_1.default.UserPlugins(submanagerOpts);
                this.link = new deps_1.default.LinkPlugins(submanagerOpts);
            }
            catch (err) {
                cli_ux_1.default.warn(err, { context: 'user/link plugins' });
            }
        }
        this.submanagers = ts_lodash_1.default.compact([this.link, this.user, this.core, this.builtin]);
    }
    async findCommand(id, options = {}) {
        const cmd = await super.findCommand(id);
        if (!cmd && options.must)
            throw new Error(`${id} not found`);
        return cmd;
    }
    get plugins() {
        const managers = ts_lodash_1.default.compact([this.link, this.user, this.core]);
        const plugins = managers.reduce((o, i) => o.concat(i.plugins), []);
        return ts_lodash_1.default.compact([...plugins, this.builtin]);
    }
    async save() {
        try {
            await this.manifest.save();
            await this.cache.save();
        }
        catch (err) {
            cli_ux_1.default.warn(err);
        }
    }
    async migrate() {
        const migrate = new deps_1.default.PluginsMigrate({
            manifest: this.manifest,
            config: this.config,
        });
        await migrate.migrate();
    }
    async getLinkedPackageName(root) {
        const pjson = await deps_1.default.file.fetchJSONFile(path.join(root, 'package.json'));
        return pjson.name;
    }
    pluginType(name) {
        const plugin = this.plugins.find(p => p.name === name);
        return plugin && plugin.type;
    }
}
exports.Plugins = Plugins;
