"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const plugin_1 = require("./plugin");
const path = require("path");
const fs = require("fs-extra");
const manager_1 = require("./manager");
const deps_1 = require("../deps");
function touch(f) {
    fs.utimesSync(f, new Date(), new Date());
}
function linkPJSON(root) {
    return deps_1.default.file.fetchJSONFile(path.join(root, 'package.json'));
}
async function getNewestJSFile(root) {
    let files = await deps_1.default.file.walk(root, {
        depthLimit: 10,
        filter: f => !['.git', 'node_modules'].includes(path.basename(f)),
    });
    return files.reduce((prev, f) => {
        if (f.stats.isDirectory())
            return prev;
        if (f.path.endsWith('.js') || f.path.endsWith('.ts')) {
            if (f.stats.mtime > prev)
                return f.stats.mtime;
        }
        return prev;
    }, new Date(0));
}
class LinkPlugins extends manager_1.PluginManager {
    async install(root) {
        this.debug('installing', root);
        const plugin = await this.loadPlugin(root, true);
        await plugin.refresh();
        await plugin.init();
    }
    async findByRoot(root) {
        await this.init();
        root = path.resolve(root);
        return this.plugins.find(p => path.resolve(p.root) === root);
    }
    async _init() {
        this.debug('init');
        const defs = await this.manifest.list('link');
        const promises = defs.map(p => this.loadPlugin(p.root));
        this.submanagers = this.plugins = await Promise.all(promises);
        if (this.plugins.length)
            this.debug('plugins:', this.plugins.map(p => p.name).join(', '));
    }
    async loadPlugin(root, forceRefresh = false) {
        const pjson = await linkPJSON(root);
        return new LinkPlugin({
            name: pjson.name,
            cache: this.cache,
            type: 'link',
            config: this.config,
            forceRefresh,
            manifest: this.manifest,
            root,
            version: pjson.version,
            pjson,
        });
    }
}
exports.LinkPlugins = LinkPlugins;
class LinkPlugin extends plugin_1.Plugin {
    constructor(opts) {
        super(opts);
        this.type = 'link';
        this.forceRefresh = opts.forceRefresh;
    }
    async resetCache() {
        await super.resetCache();
        const plugin = await this.manifestInfo();
        if (plugin)
            await this.manifest.update('link', plugin.name);
    }
    async _refresh() {
        let type = await this.refreshType();
        switch (type) {
            case 'node_modules':
                await this.updateNodeModules();
                break;
            case 'prepare':
                await this.prepare();
                break;
        }
        await super._refresh();
    }
    async _needsRefresh() {
        if (await this.refreshType())
            return true;
        return super._needsRefresh();
    }
    async refreshType() {
        if (this.forceRefresh || this.manifest.nodeVersionChanged)
            return 'node_modules';
        if (await this.updateNodeModulesNeeded())
            return 'node_modules';
        if (await this.prepareNeeded())
            return 'prepare';
    }
    async updateNodeModulesNeeded() {
        let modules = path.join(this.root, 'node_modules');
        if (!await deps_1.default.file.exists(modules))
            return true;
        let modulesInfo = await fs.stat(modules);
        let pjsonInfo = await fs.stat(path.join(this.root, 'package.json'));
        return modulesInfo.mtime < pjsonInfo.mtime;
    }
    async prepareNeeded() {
        const main = this.pjson.main;
        if (main && !await deps_1.default.file.exists(path.join(this.root, main)))
            return true;
        return (await this.lastUpdated()) < (await getNewestJSFile(this.root));
    }
    async updateNodeModules() {
        this.debug('update node modules');
        const yarn = new deps_1.default.Yarn({ config: this.config, cwd: this.root });
        await yarn.exec();
        touch(path.join(this.root, 'node_modules'));
        await this.resetCache();
    }
    async prepare() {
        this.debug('prepare');
        const { scripts } = this.pjson;
        if (!scripts || !scripts.prepare)
            return;
        const yarn = new deps_1.default.Yarn({ config: this.config, cwd: this.root });
        await yarn.exec(['run', 'prepare']);
        await this.resetCache();
    }
    async lastUpdated() {
        const plugin = await this.manifestInfo();
        if (!plugin)
            return new Date(0);
        return plugin.lastUpdated;
    }
    async manifestInfo() {
        const plugins = await this.manifest.list('link');
        const plugin = plugins.find(p => p.root === this.root);
        return plugin;
    }
}
exports.LinkPlugin = LinkPlugin;
