"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const deps_1 = require("../deps");
const heroku_cli_color_1 = require("heroku-cli-color");
const cli_ux_1 = require("cli-ux");
const util_1 = require("util");
const topic_1 = require("./topic");
const ts_lodash_1 = require("ts-lodash");
class PluginManager {
    constructor(opts) {
        this.submanagers = [];
        this._commandFinders = {};
        this.debug = require('debug')(`cli:plugins:${this.constructor.name.split('Plugin')[0].toLowerCase()}`);
        this.cacheKey = 'plugins';
        this.config = opts.config;
        this.manifest = opts.manifest || new deps_1.default.PluginManifest(this.config);
        this.cache = opts.cache || new deps_1.default.PluginCache(this.config);
    }
    init() {
        if (this._initPromise)
            return this._initPromise;
        return (this._initPromise = (async () => {
            await this._init();
            await this.initSubmanagers();
        })());
    }
    get needsRefresh() {
        if (this._needsRefreshPromise)
            return this._needsRefreshPromise;
        return (this._needsRefreshPromise = (async () => {
            if (await this._needsRefresh())
                return true;
            const errHandle = (err) => cli_ux_1.default.warn(err, { context: `needsRefresh: ${this.constructor.name}` });
            const promises = this.submanagers.map(m => m.needsRefresh.catch(errHandle));
            for (let s of promises)
                if (await s)
                    return true;
            return false;
        })());
    }
    async _needsRefresh() {
        return false;
    }
    refresh() {
        if (this._refreshPromise)
            return this._refreshPromise;
        return (this._refreshPromise = (async () => {
            this.debug('refresh');
            await this._refresh();
            const errHandle = (err) => cli_ux_1.default.warn(err, { context: `refresh: ${this.constructor.name}` });
            const promises = this.submanagers.map(async (m) => {
                if (await m.needsRefresh)
                    return m.refresh().catch(errHandle);
            });
            for (let s of promises)
                await s;
        })());
    }
    async _refresh() { }
    findTopic(name) {
        return topic_1.Topic.findTopic(name, this.topics);
    }
    get load() {
        if (!this._loadPromise) {
            this._loadPromise = (async () => {
                await this._load();
                const errHandle = (err) => cli_ux_1.default.warn(err, { context: `load: ${this.constructor.name}` });
                const promises = this.submanagers.map(m => m.load.catch(errHandle));
                for (let s of promises)
                    await s;
                this.commandIDs = await this.fetchCommandIDs();
                this.topics = await this.fetchTopics();
                this.rootCommands = await this.fetchRootCommands();
                this.aliases = await this.fetchAliases();
            })();
        }
        return this._loadPromise;
    }
    async _load() { }
    async fetchCommandIDs() {
        const fetch = () => this._commandIDs();
        let ids = await this.cache.fetch(this.cacheKey, 'command:ids', fetch);
        return ts_lodash_1.default.uniq(ids.concat(...this.submanagers.map(s => s.commandIDs)).sort());
    }
    findCommand(id) {
        if (this._commandFinders[id])
            return this._commandFinders[id];
        return (this._commandFinders[id] = (async () => {
            if (!this.commandIDs.includes(id))
                return;
            let cmd = await this._findCommand(await this.unalias(id));
            if (cmd)
                return cmd;
            const errHandle = (err) => cli_ux_1.default.warn(err, { context: `findCommand: ${this.constructor.name}` });
            const promises = this.submanagers.map(s => s.findCommand(id).catch(errHandle));
            for (let p of promises) {
                let cmd = await p;
                if (cmd)
                    return cmd;
            }
        })());
    }
    findCommandInfo(id) {
        let cmd = this.rootCommands[id];
        if (cmd)
            return cmd;
        let topic = topic_1.Topic.findTopic(topic_1.Topic.parentTopicIDof(id), this.topics);
        if (!topic)
            return;
        return topic.commands[id];
    }
    require(p, id) {
        this.debug('Reading command %s at %s', id, p);
        let Command;
        try {
            Command = deps_1.default.util.undefault(require(p));
        }
        catch (err) {
            cli_ux_1.default.warn(err, { context: `Error reading command from ${p}` });
        }
        if (!Command || !Command._version) {
            let extra = deps_1.default.util.isEmpty(Command)
                ? 'Does the command have `export default class extends Command {...}`?'
                : `Received: ${util_1.inspect(Command)}`;
            throw new Error(`${p} does not appear to be a valid command.\n${extra}`);
        }
        return Command;
    }
    async _init() { }
    async _topics() {
        return {};
    }
    async _commandIDs() {
        return [];
    }
    async _aliases() {
        return {};
    }
    async _findCommand(_) {
        return undefined;
    }
    async unalias(id) {
        const aliases = Object.entries(await this._aliases());
        const alias = aliases.find(([, aliases]) => aliases.includes(id));
        return alias ? alias[0] : id;
    }
    async initSubmanagers() {
        const errHandle = (err) => cli_ux_1.default.warn(err, { context: `init: ${this.constructor.name}` });
        const promises = this.submanagers.map(m => m.init().catch(errHandle));
        for (let s of promises)
            await s;
    }
    async addCommandsToTopics(topics) {
        for (let command of await this._commandIDs()) {
            const topicID = topic_1.Topic.parentTopicIDof(command);
            if (!topicID)
                continue;
            const info = await this.fetchCommandInfo(command);
            let topic = await topic_1.Topic.findOrCreateTopic({ name: topicID }, topics);
            topic.commands[command] = info;
        }
    }
    async findCommandHelpLine(id) {
        const c = await this.findCommand(id);
        if (!c)
            throw new Error('command not found');
        return c.buildHelpLine(this.config);
    }
    async findCommandHelp(id) {
        const c = await this.findCommand(id);
        if (!c)
            throw new Error('command not found');
        let help = c.buildHelp(this.config);
        if (!heroku_cli_color_1.color.supportsColor)
            help = deps_1.default.stripAnsi(help);
        return help;
    }
    async fetchTopics() {
        const fetch = async () => {
            let topics = await this._topics();
            await this.addCommandsToTopics(topics);
            return topics;
        };
        let topics = await this.cache.fetch(this.cacheKey, 'topics', fetch);
        return topic_1.Topic.mergeSubtopics(topics, ...this.submanagers.map(m => m.topics));
    }
    async fetchRootCommands() {
        const fetch = async () => await this._rootCommands();
        let rootCommands = await this.cache.fetch(this.cacheKey, 'root_commands', fetch);
        for (let s of this.submanagers) {
            rootCommands = Object.assign({}, rootCommands, s.rootCommands);
        }
        return rootCommands;
    }
    async _rootCommands() {
        const commands = {};
        for (let command of await this._commandIDs()) {
            const topicID = topic_1.Topic.parentTopicIDof(command);
            if (topicID)
                continue;
            const info = await this.fetchCommandInfo(command);
            commands[command] = info;
        }
        return commands;
    }
    async fetchCommandInfo(id) {
        let cmd = await this.findCommand(id);
        if (!cmd)
            throw new Error(`${id} not found`);
        return {
            id,
            hidden: cmd.hidden,
            help: await this.findCommandHelp(id),
            helpLine: await this.findCommandHelpLine(id),
        };
    }
    async fetchAliases() {
        const fetch = () => this._aliases();
        let aliases = await this.cache.fetch(this.cacheKey, 'aliases', fetch);
        for (let s of this.submanagers) {
            for (let [k, v] of Object.entries(s.aliases || {})) {
                aliases[k] = ts_lodash_1.default.uniq([...(aliases[k] || []), ...deps_1.default.util.toArray(v)]);
            }
        }
        return aliases;
    }
}
exports.PluginManager = PluginManager;
