"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const deps_1 = require("../deps");
const topic_1 = require("./topic");
const manager_1 = require("./manager");
const path = require("path");
class Plugin extends manager_1.PluginManager {
    constructor(options) {
        super(options);
        if (options.pjson)
            this.pjson = options.pjson;
        this.name = options.name;
        this.version = options.version;
        this.root = options.root;
        this.cacheKey = [options.name, options.type, options.version].join(':');
        this.debug = require('debug')(`cli:plugins:${this.cacheKey}`);
    }
    async _init() {
        if (this.type !== 'builtin') {
            this.pjson = this.pjson || (await deps_1.default.file.readJSON(path.join(this.root, 'package.json')));
            if (!this.pjson['cli-engine'])
                this.pjson['cli-engine'] = {};
            this.name = this.pjson.name;
            this.version = this.pjson.version;
        }
    }
    async _refresh() {
        await this.resetCache();
    }
    async _needsRefresh() {
        return !await this.cache.get(this.cacheKey, 'topics');
    }
    _commandIDs() {
        return deps_1.default.util.concatPromiseArrays([this.commandIDsFromDir(), this.fetchCommandIDsFromModule()]);
    }
    async _aliases() {
        return deps_1.default.util.objValsToArrays(this.pjson['cli-engine'].aliases);
    }
    async _findCommand(id) {
        let cmd = await this.findCommandInModule(id);
        if (!cmd)
            cmd = await this.findCommandInDir(id);
        if (cmd) {
            cmd = this.addPluginToCommand(cmd);
            return cmd;
        }
    }
    addPluginToCommand(cmd) {
        cmd.plugin = {
            type: this.type,
            root: this.root,
            name: this.name,
            version: this.version,
        };
        return cmd;
    }
    async _topics() {
        return topic_1.Topic.mergeSubtopics(await this.fetchTopicsFromModule());
    }
    async resetCache() {
        await this.cache.reset(this.cacheKey);
    }
    async findCommandInModule(id) {
        const m = await this.fetchModule();
        if (!m)
            return;
        return m.commands.find(c => c.id === id);
    }
    async findCommandInDir(id) {
        const dir = this.commandsDir;
        if (!dir)
            return;
        let p;
        try {
            p = this.commandPath(id);
        }
        catch (err) {
            if (err.code === 'MODULE_NOT_FOUND')
                return;
            throw err;
        }
        if (!await deps_1.default.file.exists(p))
            return;
        return this.require(p, id);
    }
    commandPath(id) {
        if (!this.commandsDir)
            throw new Error('commandsDir not set');
        return require.resolve(path.join(this.commandsDir, id.split(':').join(path.sep)));
    }
    async fetchModule() {
        if (this._module)
            return this._module;
        return (this._module = (async () => {
            if (!this.pjson.main)
                return;
            this.debug(`requiring ${this.name}@${this.version}`);
            const m = Object.assign({ commands: [], topics: [] }, require(path.join(this.root, this.pjson.main)));
            if (m.topic)
                m.topics.push(m.topic);
            m.commands = m.commands.map(deps_1.default.util.undefault);
            const hooks = new deps_1.default.Hooks(this.config);
            await hooks.run('plugins:parse', { module: m, pjson: this.pjson });
            let legacy = new deps_1.default.PluginLegacy(this.config);
            return legacy.convert(m);
        })());
    }
    async fetchCommandIDsFromModule() {
        const m = await this.fetchModule();
        if (!m)
            return [];
        return m.commands.map(m => m.id);
    }
    async fetchTopicsFromModule() {
        const m = await this.fetchModule();
        if (!m)
            return {};
        return m.topics.reduce((topics, t) => {
            topics[t.name] = new topic_1.Topic(t);
            return topics;
        }, {});
    }
    commandIDsFromDir() {
        const d = this.commandsDir;
        if (!d)
            return Promise.resolve([]);
        return new Promise((resolve, reject) => {
            let ids = [];
            deps_1.default
                .klaw(d, { depthLimit: 10 })
                .on('data', f => {
                if (!f.stats.isDirectory() && f.path.endsWith('.js')) {
                    let parsed = path.parse(f.path);
                    let p = path.relative(d, path.join(parsed.dir, parsed.name));
                    ids.push(p.split(path.sep).join(':'));
                }
            })
                .on('error', reject)
                .on('end', () => resolve(ids));
        });
    }
    get commandsDir() {
        let d = this.pjson['cli-engine'].commandsDir;
        if (d)
            return path.join(this.root, d);
    }
}
exports.Plugin = Plugin;
