"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const deps_1 = require("./deps");
const ts_lodash_1 = require("ts-lodash");
const path = require("path");
const cli_ux_1 = require("cli-ux");
const debug = require('debug')('cli:updater');
async function mtime(f) {
    const { mtime } = await deps_1.default.file.stat(f);
    return deps_1.default.moment(mtime);
}
function timestamp(msg) {
    return `[${deps_1.default.moment().format()}] ${msg}`;
}
class Updater {
    constructor(config) {
        this.config = config;
        this.lock = new deps_1.default.Lock(config, `${this.autoupdatefile}.lock`);
    }
    get autoupdatefile() {
        return path.join(this.config.cacheDir, 'autoupdate');
    }
    get autoupdatelogfile() {
        return path.join(this.config.cacheDir, 'autoupdate.log');
    }
    get versionFile() {
        return path.join(this.config.cacheDir, `${this.config.channel}.version`);
    }
    get clientRoot() {
        return path.join(this.config.dataDir, 'client');
    }
    get clientBin() {
        let b = path.join(this.clientRoot, 'bin', this.config.bin);
        return this.config.windows ? `${b}.cmd` : b;
    }
    get binPath() {
        if (!this._binPath)
            this._binPath = (async () => {
                if (!this.config.updateDisabled && (await deps_1.default.file.exists(this.clientBin))) {
                    return this.clientBin;
                }
                return process.env.CLI_BINPATH || this.config.bin;
            })();
        return this._binPath;
    }
    s3url(channel, p) {
        if (!this.config.s3.host)
            throw new Error('S3 host not defined');
        return `https://${this.config.s3.host}/${this.config.name}/channels/${channel}/${p}`;
    }
    async fetchManifest(channel) {
        try {
            let { body } = await deps_1.default.HTTP.get(this.s3url(channel, `${this.config.platform}-${this.config.arch}`));
            return body;
        }
        catch (err) {
            if (err.statusCode === 403)
                throw new Error(`HTTP 403: Invalid channel ${channel}`);
            throw err;
        }
    }
    async fetchVersion(download) {
        let v;
        try {
            if (!download)
                v = await deps_1.default.file.readJSON(this.versionFile);
        }
        catch (err) {
            if (err.code !== 'ENOENT')
                throw err;
        }
        if (!v) {
            debug('fetching latest %s version', this.config.channel);
            let { body } = await deps_1.default.HTTP.get(this.s3url(this.config.channel, 'version'));
            v = body;
            await this._catch(() => deps_1.default.file.outputJSON(this.versionFile, v));
        }
        return v;
    }
    async _catch(fn) {
        try {
            return await Promise.resolve(fn());
        }
        catch (err) {
            debug(err);
        }
    }
    async update(manifest) {
        const downgrade = await this.lock.write();
        let base = this.base(manifest);
        const filesize = require('filesize');
        if (!this.config.s3.host)
            throw new Error('S3 host not defined');
        let url = `https://${this.config.s3.host}/${this.config.name}/channels/${manifest.channel}/${base}.tar.gz`;
        let { response: stream } = await deps_1.default.HTTP.stream(url);
        let output = path.join(this.clientRoot, manifest.version);
        await this._mkdirp(this.clientRoot);
        await this._remove(output);
        if (cli_ux_1.cli.action.frames) {
            // if spinner action
            let total = stream.headers['content-length'];
            let current = 0;
            const updateStatus = ts_lodash_1.default.throttle((newStatus) => {
                cli_ux_1.cli.action.status = newStatus;
            }, 500, { leading: true, trailing: false });
            stream.on('data', data => {
                current += data.length;
                updateStatus(`${filesize(current)}/${filesize(total)}`);
            });
        }
        await this.extract(stream, this.clientRoot, manifest.sha256gz);
        await this._rename(path.join(this.clientRoot, base), output);
        await this._createBin(path.join(output, 'bin', this.config.bin), manifest);
        await downgrade();
    }
    extract(stream, dir, sha) {
        const zlib = require('zlib');
        const tar = require('tar-fs');
        const crypto = require('crypto');
        return new Promise((resolve, reject) => {
            let shaValidated = false;
            let extracted = false;
            let check = () => {
                if (shaValidated && extracted) {
                    resolve();
                }
            };
            let fail = (err) => {
                this._remove(dir).then(() => reject(err));
            };
            let hasher = crypto.createHash('sha256');
            stream.on('error', fail);
            stream.on('data', d => hasher.update(d));
            stream.on('end', () => {
                let shasum = hasher.digest('hex');
                if (sha === shasum) {
                    shaValidated = true;
                    check();
                }
                else {
                    reject(new Error(`SHA mismatch: expected ${shasum} to be ${sha}`));
                }
            });
            let ignore = function (_, header) {
                switch (header.type) {
                    case 'directory':
                    case 'file':
                        debug(header.name);
                        return false;
                    case 'symlink':
                        return true;
                    default:
                        throw new Error(header.type);
                }
            };
            let extract = tar.extract(dir, { ignore });
            extract.on('error', fail);
            extract.on('finish', () => {
                extracted = true;
                check();
            });
            let gunzip = zlib.createGunzip();
            gunzip.on('error', fail);
            stream.pipe(gunzip).pipe(extract);
        });
    }
    async _rename(from, to) {
        await deps_1.default.file.rename(from, to);
    }
    async _remove(dir) {
        if (await deps_1.default.file.exists(dir)) {
            await deps_1.default.file.remove(dir);
        }
    }
    async _mkdirp(dir) {
        await deps_1.default.file.mkdirp(dir);
    }
    base(manifest) {
        return `${this.config.name}-v${manifest.version}-${this.config.platform}-${this.config.arch}`;
    }
    async autoupdateNeeded() {
        try {
            const m = await mtime(this.autoupdatefile);
            return m.isBefore(deps_1.default.moment().subtract(5, 'hours'));
        }
        catch (err) {
            if (err.code !== 'ENOENT')
                console.error(err.stack);
            debug('autoupdate ENOENT');
            return true;
        }
    }
    async autoupdate(force = false) {
        try {
            await this.warnIfUpdateAvailable();
            if (!force && !await this.autoupdateNeeded())
                return;
            debug('autoupdate running');
            await deps_1.default.file.outputFile(this.autoupdatefile, '');
            const binPath = await this.binPath;
            if (!binPath) {
                debug('no binpath set');
                return;
            }
            debug(`spawning autoupdate on ${binPath}`);
            let fd = await deps_1.default.file.open(this.autoupdatelogfile, 'a');
            deps_1.default.file.write(fd, timestamp(`starting \`${binPath} update --autoupdate\` from ${process.argv.slice(2, 3).join(' ')}\n`));
            this.spawnBinPath(binPath, ['update', '--autoupdate'], {
                detached: !this.config.windows,
                stdio: ['ignore', fd, fd],
                env: this.autoupdateEnv,
            })
                .on('error', (e) => cli_ux_1.cli.warn(e, { context: 'autoupdate:' }))
                .unref();
        }
        catch (e) {
            cli_ux_1.cli.warn(e, { context: 'autoupdate:' });
        }
    }
    get timestampEnvVar() {
        // TODO: use function from cli-engine-config
        let bin = this.config.bin.replace('-', '_').toUpperCase();
        return `${bin}_TIMESTAMPS`;
    }
    get skipAnalyticsEnvVar() {
        let bin = this.config.bin.replace('-', '_').toUpperCase();
        return `${bin}_SKIP_ANALYTICS`;
    }
    get autoupdateEnv() {
        return Object.assign({}, process.env, {
            [this.timestampEnvVar]: '1',
            [this.skipAnalyticsEnvVar]: '1',
        });
    }
    async warnIfUpdateAvailable() {
        await this._catch(async () => {
            if (!this.config.s3)
                return;
            let v = await this.fetchVersion(false);
            let local = this.config.version.split('.');
            let remote = v.version.split('.');
            if (parseInt(local[0]) < parseInt(remote[0]) || parseInt(local[1]) < parseInt(remote[1])) {
                cli_ux_1.cli.warn(`${this.config.name}: update available from ${this.config.version} to ${v.version}`);
            }
            if (v.message) {
                cli_ux_1.cli.warn(`${this.config.name}: ${v.message}`);
            }
        });
    }
    spawnBinPath(binPath, args, options) {
        debug(binPath, args);
        if (this.config.windows) {
            args = ['/c', binPath, ...args];
            return deps_1.default.crossSpawn(process.env.comspec || 'cmd.exe', args, options);
        }
        else {
            return deps_1.default.crossSpawn(binPath, args, options);
        }
    }
    async _createBin(dst, manifest) {
        let src = this.clientBin;
        let body;
        if (this.config.windows) {
            body = `@echo off
"%~dp0\\..\\${manifest.version}\\bin\\heroku.cmd" %*
`;
        }
        else {
            body = `#!/usr/bin/env bash
if [ "$DEBUG" == "*" ]; then
  echo "running ${dst}" "$@"
fi
"${dst}" "$@"
`;
        }
        await deps_1.default.file.outputFile(src, body, { mode: 0o777 });
    }
}
exports.Updater = Updater;
